/*
 * Copyright (c) 2007 Markus Fisch <mf@markusfisch.de>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Markus Fisch nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#include "Application.h"
#include "PieMenuWindow.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <errno.h>
#include <pwd.h>
#include <libgen.h>
#include <sstream>

using namespace PieDock;

/**
 * Initialize application
 *
 * @param argc - number of arguments
 * @param argv - arguments
 */
Application::Application( int argc, char **argv ) :
	display( XOpenDisplay( 0 ) ),
	root( DefaultRootWindow( display ) ),
	settings( getHome()+std::string( "/." )+
		basename( *argv )+std::string( "rc" ) ),
	suspend( StandBy )
{
	if(	!display )
		throw "Cannot to open display";

	std::ostringstream s;

	s << getHome() << "/." << basename( *argv ) << 
		std::string( "-" ) << DisplayString( display );

	socketFile = s.str();
}

/**
 * Clean up
 */
Application::~Application()
{
	XCloseDisplay( display );
}

/**
 * Try to wake already existing instance; returns true if successful
 */
bool Application::wake() const
{
	struct stat buf;

	if( stat( socketFile.c_str(), &buf ) < 0 )
		return false;

	struct sockaddr_un address;
	int s;

	if( (s = socket( PF_UNIX, SOCK_DGRAM, 0 )) < 0 )
		throw strerror( errno );

	memset( &address, sizeof( struct sockaddr_un ), 0 );

	address.sun_family = AF_LOCAL;
	strncpy( 
		reinterpret_cast<char *>( address.sun_path ), 
		socketFile.c_str(), 
		UnixPathMax );

	if( connect( s, 
		(struct sockaddr *) &address, 
		sizeof( struct sockaddr_un ) ) < 0 )
	{
		// if there's no listener assume the file has been left
		// over from a previous instance and try to remove it
		// to start anew
		if( errno != ECONNREFUSED ||
			unlink( socketFile.c_str() ) )
			throw strerror( errno );

		return false;
	}

	char command = WakeUp;

	if( send( s, &command, sizeof( char ), 0 ) < 0 )
		throw strerror( errno );
	
	close( s );

	return true;
}

/**
 * Run application
 *
 * @param stopFlag - pointer to stop flag
 */
int Application::run( bool *stopFlag )
{
	int xfd = ConnectionNumber( display );
	int s = 0;

	// at first, load settings
	settings.load();

	// create socket for external activation
	{
		struct sockaddr_un address;

		if( (s = socket( PF_UNIX, SOCK_DGRAM, 0 )) < 0 )
			throw strerror( errno );

		memset( &address, sizeof( struct sockaddr_un ), 0 );

		address.sun_family = AF_LOCAL;
		strncpy( 
			reinterpret_cast<char *>( address.sun_path ), 
			socketFile.c_str(), 
			UnixPathMax );

		if( bind( s, 
			(struct sockaddr *) &address, 
			sizeof( struct sockaddr_un ) ) < 0 )
			throw strerror( errno );
	}

	grabTriggers();

	for( PieMenuWindow w( *this ); !*stopFlag; )
	{
		if( !XPending( display ) )
		{
			fd_set rfds;
			struct timeval tv, *ptv = 0;

			FD_ZERO( &rfds );
			FD_SET( xfd, &rfds );
			FD_SET( s, &rfds );

			if( suspend > 0 )
			{
				tv.tv_sec = 0;
				tv.tv_usec = suspend;
				ptv = &tv;
			}

			// wait for descriptors to become readable
			{
				int highest = (s > xfd ? s : xfd)+1;
				int hits;

				if( (hits = select( highest, &rfds, 0, 0, ptv )) < 0 )
				{
					// signal caught
					if( errno == EINTR )
						continue;

					// some descriptor has closed
					break;
				}
				else if( !hits )
				{
					// timeout
					if( suspend == Active )
						w.draw();

					continue;
				}
				else 
				{
					// some descriptor has become readable
					if( FD_ISSET( s, &rfds ) )
					{
						char byte = 0;

						recv( s, &byte, 1, 0 );

						if( byte == WakeUp &&
							suspend == StandBy )
						{
							// why ? see below
							ungrabTriggers();
							w.appear();
							suspend = Active;
						}
					}

					if( !FD_ISSET( xfd, &rfds ) )
						continue;
				}
			}
		}

		XEvent event;

		bzero( &event, sizeof( event ) );
		XNextEvent( display, &event );

		if( suspend == StandBy &&
			event.xany.window == root && 
			(event.type == ButtonRelease ||
				event.type == KeyPress) )
		{
			// ungrab triggers to avoid being triggered again while
			// already being visible; since receiving a new trigger 
			// event will be preceded by a LeaveWindowEvent in 
			// PieMenuWindow which will cause the menu to disappear 
			// again, it's best to simply ungrab the triggers
			ungrabTriggers();
			w.appear();
			suspend = Active;
		}
		else if( suspend == Active &&
			event.xany.window == w.getWindow() &&
			!w.processEvent( event ) )
		{
			suspend = StandBy;
			grabTriggers();
		}
	}

	ungrabTriggers();
	close( s );
	unlink( socketFile.c_str() );

	return 0;
}

/**
 * Return home directory
 */
std::string Application::getHome()
{
	const char *home;

	if( (home = getenv( "HOME" )) )
		return home;

	struct passwd *pw = getpwuid( getuid() );

    if( pw )
		return pw->pw_dir;

	return "";
}

/**
 * Grab triggers
 */
void Application::grabTriggers()
{
	// buttons
	for( Settings::Buttons::iterator i = settings.getButtons().begin();
		i != settings.getButtons().end();
		i++ )
		XGrabButton(
			display, 
			(*i).button, 
			(*i).modifier, 
			root, 
			True, 
			ButtonReleaseMask, 
			GrabModeAsync,
			GrabModeAsync, 
			None, 
			None );

	// keys
	for( Settings::Keys::iterator i = settings.getKeys().begin();
		i != settings.getKeys().end();
		i++ )
		XGrabKey(
			display, 
			XKeysymToKeycode( display, (*i).keySym ), 
			(*i).modifier, 
			root,
			True, 
			GrabModeAsync, 
			GrabModeAsync );
}

/**
 * Ungrab triggers
 */
void Application::ungrabTriggers()
{
	// buttons
	for( Settings::Buttons::iterator i = settings.getButtons().begin();
		i != settings.getButtons().end();
		i++ )
		XUngrabButton(
			display, 
			(*i).button, 
			(*i).modifier, 
			root );

	// keys
	for( Settings::Keys::iterator i = settings.getKeys().begin();
		i != settings.getKeys().end();
		i++ )
		XUngrabKey(
			display, 
			XKeysymToKeycode( display, (*i).keySym ), 
			(*i).modifier, 
			root );
}
