/*
 * Copyright (c) 2007 Markus Fisch <mf@markusfisch.de>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Markus Fisch nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#include "Blender.h"

using namespace PieDock;

/**
 * Initialize blender
 *
 * @param c - canvas image of arbitrary color-depth
 */
Blender::Blender( Surface &c ) :
	canvas( &c )
{
}

/**
 * Blend RGBA surface into canvas
 *
 * @param src - RGBA surface to blend into canvas
 * @param x - left position in canvas (optional)
 * @param y - upper position in canvas (optional)
 * @param a - alpha value (optional)
 */
void Blender::blend( const RgbaSurface &src, int x, int y, int a )
{
	Details details = {
		canvas->getData(),
		src.getData(),
		src.getWidth(),
		src.getHeight(),
		src.getPadding(),
		0,
		a };

	if( x > canvas->getWidth() ||
		y > canvas->getHeight() )
		return;

	if( x < 0 )
	{
		int o = x*-src.getBytesPerPixel();

		details.src += o;
		details.srcSkip += o;
		details.length += x;

		x = 0;
	}

	if( y < 0 )
	{
		details.src += y*-src.getBytesPerLine();
		details.repeats += y;
		y = 0;
	}

	details.destSkip = x*canvas->getBytesPerPixel();
	details.dest += 
		y*canvas->getBytesPerLine()+
		details.destSkip;
	details.destSkip += canvas->getPadding();

	{
		int e = x+details.length;

		if( e >= canvas->getWidth() )
		{
			details.length = canvas->getWidth()-x;
			details.srcSkip += 
				(e-canvas->getWidth())*
				src.getBytesPerPixel();
		}
		else
			details.destSkip += 
				(canvas->getWidth()-e)*
				canvas->getBytesPerPixel();
	}

	if( y+details.repeats >= canvas->getHeight() )
		details.repeats = canvas->getHeight()-y;

	if( details.length <= 0 ||
		details.repeats <= 0 )
		return;

	switch( canvas->getDepth() )
	{
		case 32:
			blendInto32Bit( details );
			break;
		case 24:
			blendInto24Bit( details );
			break;
		case 16:
			blendInto16Bit( details );
			break;
		default:
			throw "Pixel depth not supported !";
	}
}

/**
 * Blend into RGBA image
 *
 * @param details - blending details
 */
void Blender::blendInto32Bit( Details &details )
{
	unsigned char *src = details.src;
	unsigned char *dest = details.dest;

	if( details.alpha != 0xff )
	{
		double alphaMax = ((double) 255*((double) 255/(double) details.alpha));

		for( int r = details.repeats; 
			r--; 
			dest += details.destSkip, src += details.srcSkip )
			for( int l = details.length; l--; )
				if( !(*((int *) src) & 0xff000000) )
				{
					src += 4;
					dest += 4;
				}
				else
				{
					double diff;
					double mod = alphaMax/(double) (*(src+3));

					// it must be done that brain-dead, if you write it in 
					// just one line, things get messed up terribly by the 
					// compiler (at least up to version 3.3)
					register int red = *(src++);
					register int green = *(src++);
					register int blue = *(src++);

					diff = red;
					diff -= *dest;
					diff /= mod;
					*(dest++) += (int) diff;

					diff = green;
					diff -= *dest;
					diff /= mod;
					*(dest++) += (int) diff;

					diff = blue;
					diff -= *dest;
					diff /= mod;
					*(dest++) += (int) diff;

					src++;
					dest++;
				}

		return;
	}

	for( int r = details.repeats; 
		r--; 
		dest += details.destSkip, src += details.srcSkip )
		for( int l = details.length; l--; )
			switch( *((int *) src) & 0xff000000 )
			{
				case 0:
					src += 4;
					dest += 4;
					break;
				case 0xff000000:
					*((int *) dest) = *((int *) src);
					src += 4;
					dest += 4;
					break;
				default:
					{
						double diff;
						double mod = (double) 255/(double) (*(src+3));

						diff = *(src++);
						diff -= *dest;
						diff /= mod;
						*(dest++) += (int) diff;

						diff = *(src++);
						diff -= *dest;
						diff /= mod;
						*(dest++) += (int) diff;

						diff = *(src++);
						diff -= *dest;
						diff /= mod;
						*(dest++) += (int) diff;

						src++;
						dest++;
					}
					break;
			}
}

/**
 * Blend into RGB image
 *
 * @param details - blending details
 */
void Blender::blendInto24Bit( Details &details )
{
	unsigned char *src = details.src;
	unsigned char *dest = details.dest;

	if( details.alpha != 0xff )
	{
		double alphaMax = ((double) 255*((double) 255/(double) details.alpha));

		for( int r = details.repeats; 
			r--; 
			dest += details.destSkip, src += details.srcSkip )
			for( int l = details.length; l--; )
				if( !(*((int *) src) & 0xff000000) )
				{
					src += 4;
					dest += 3;
				}
				else
				{
					double diff;
					double mod = alphaMax/(double) (*(src+3));

					register int red = *(src++);
					register int green = *(src++);
					register int blue = *(src++);

					diff = red;
					diff -= *dest;
					diff /= mod;
					*(dest++) += (int) diff;

					diff = green;
					diff -= *dest;
					diff /= mod;
					*(dest++) += (int) diff;

					diff = blue;
					diff -= *dest;
					diff /= mod;
					*(dest++) += (int) diff;

					src++;
				}

		return;
	}

	for( int r = details.repeats; 
		r--; 
		dest += details.destSkip, src += details.srcSkip )
		for( int l = details.length; l--; )
			switch( *((int *) src) & 0xff000000 )
			{
				case 0:
					src += 4;
					dest += 3;
					break;
				case 0xff000000:
					*(dest++) = *(src++);
					*(dest++) = *(src++);
					*(dest++) = *(src++);
					src++;
					break;
				default:
					{
						double diff;
						double mod = (double) 255/(double) (*(src+3));

						diff = *(src++);
						diff -= *dest;
						diff /= mod;
						*(dest++) += (int) diff;

						diff = *(src++);
						diff -= *dest;
						diff /= mod;
						*(dest++) += (int) diff;

						diff = *(src++);
						diff -= *dest;
						diff /= mod;
						*(dest++) += (int) diff;

						src++;
					}
					break;
			}
}

/**
 * Blend into HighColor image
 *
 * @param details - blending details
 */
void Blender::blendInto16Bit( Details &details )
{
	unsigned char *src = details.src;
	unsigned char *dest = details.dest;

	if( details.alpha != 0xff )
	{
		double alphaMax = ((double) 255*((double) 255/(double) details.alpha));

		for( int r = details.repeats; 
			r--; 
			dest += details.destSkip, src += details.srcSkip )
			for( int l = details.length; l--; )
				if( !(*((int *) src) & 0xff000000) )
				{
					src += 4;
					dest += 2;
				}
				else
				{
					double diff;
					double mod = alphaMax/(double) (*(src+3));

					register int red = *(src++);
					register int green = *(src++);
					register int blue = *(src++);

					short pixel = *((short *) dest);

					blue = (pixel<<3)&0xf8;
					green = (pixel>>3)&0xf8;
					red = (pixel>>8)&0xf8;

					diff = red;
					diff -= blue;
					diff /= mod;
					blue += (int) diff;

					diff = green;
					diff -= green;
					diff /= mod;
					green += (int) diff;

					diff = blue;
					diff -= red;
					diff /= mod;
					red += (int) diff;

					*((short *) dest) = 
						((short) ((blue&0xf8)>>3)) |
						((short) ((green&0xf8)<<3)) |
						((short) ((red&0xf8)<<8));

					dest += 2;
					src++;
				}

		return;
	}

	for( int r = details.repeats; 
		r--; 
		dest += details.destSkip, src += details.srcSkip )
		for( int l = details.length; l--; )
			switch( *((int *) src) & 0xff000000 )
			{
				case 0:
					src += 4;
					dest += 2;
					break;
				case 0xff000000:
					{
						register int blue = *(src++);
						register int green = *(src++);
						register int red = *(src++);

						*((short *) dest) = 
							((short) ((blue&0xf8)>>3)) |
							((short) ((green&0xf8)<<3)) |
							((short) ((red&0xf8)<<8));

						src++;
						dest += 2;
					}
					break;
				default:
					{
						double diff;
						double mod = (double) 255/(double) (*(src+3));

						short pixel = *((short *) dest);

						register int blue = (pixel<<3)&0xf8;
						register int green = (pixel>>3)&0xf8;
						register int red = (pixel>>8)&0xf8;

						diff = *(src++);
						diff -= blue;
						diff /= mod;
						blue += (int) diff;

						diff = *(src++);
						diff -= green;
						diff /= mod;
						green += (int) diff;

						diff = *(src++);
						diff -= red;
						diff /= mod;
						red += (int) diff;

						*((short *) dest) = 
							((short) ((blue&0xf8)>>3)) |
							((short) ((green&0xf8)<<3)) |
							((short) ((red&0xf8)<<8));

						dest += 2;
						src++;
					}
					break;
			}
}
