/*
 * Copyright (c) 2007 Markus Fisch <mf@markusfisch.de>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Markus Fisch nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#include "Menu.h"
#include "WindowManager.h"

#include <X11/Xutil.h>
#include <unistd.h>

using namespace PieDock;

/**
 * Initialize menu
 *
 * @param app - application
 */
Menu::Menu( Application *a ) :
	app( a ),
	selected( 0 )
{
	// create menu items for sticky icons
	Settings::Icons stickyIcons = app->getSettings().getStickyIcons();

	for( Settings::Icons::iterator i = stickyIcons.begin();
		i != stickyIcons.end();
		i++ )
		menuItemsMap[(*i)] = new MenuItem( (*i), true );
}

/**
 * Clean up
 */
Menu::~Menu()
{
	for( MenuItemsMap::iterator i = menuItemsMap.begin();
		i != menuItemsMap.end();
		i++ )
		delete (*i).second;

	menuItemsMap.clear();
}

/**
 * Build menu
 */
void Menu::update()
{
	// clear menu items vector
	menuItemsVector.clear();

	// add sticky icons to menu item vector
	{
		Settings::Icons stickyIcons = app->getSettings().getStickyIcons();

		for( Settings::Icons::iterator i = stickyIcons.begin();
			i != stickyIcons.end();
			i++ )
			menuItemsVector.push_back( menuItemsMap[(*i)] );
	}

	// clear all windows of existing menuItems
	for( MenuItemsMap::iterator i = menuItemsMap.begin();
		i != menuItemsMap.end();
		i++ )
		(*i).second->clearWindows();

	// assign windows to menu items
	{
		IconMap *iconMap = &app->getSettings().getIconMap();

		// get window list
		WindowManager::WindowList wl( app->getDisplay() );

		for( WindowManager::WindowList::iterator i = wl.begin();
			i != wl.end();
			i++ )
		{
			XClassHint xch;

			if( !XGetClassHint( app->getDisplay(), (*i), &xch ) )
				continue;

			Icon *icon;

			if( !(icon = iconMap->getIconByTitle(
					WindowManager::getTitle( app->getDisplay(), (*i) ) )) &&
				!(icon = iconMap->getIconByClass( xch.res_class )) &&
				!(icon = iconMap->getIconByName( xch.res_name )) )
				icon = iconMap->getFallbackIcon( xch.res_name );

			MenuItemsMap::iterator m;

			if( (m = menuItemsMap.find( icon )) != menuItemsMap.end() )
			{
				if( !(*m).second->isSticky() &&
					!(*m).second->hasWindows() )
					menuItemsVector.push_back( (*m).second );

				(*m).second->addWindow( app->getDisplay(), (*i) );
			}
			else
			{
				MenuItem *item = new MenuItem( icon );

				item->addWindow( app->getDisplay(), (*i) );
				menuItemsMap[icon] = item;
				menuItemsVector.push_back( item );
			}
		}
	}

	// remove all menu items that have no windows and
	// are not sticky
	for( MenuItemsMap::iterator i = menuItemsMap.begin();
		i != menuItemsMap.end();
		i++ )
		if( !(*i).second->isSticky() &&
			!(*i).second->hasWindows() )
			menuItemsMap.erase( i );
}

/**
 * Execute some action for the seleced icon
 *
 * @param a - action to execute (optional)
 */
void Menu::execute( Settings::Action a )
{
	if( !selected )
		return;

	// if there are no windows only Launch is allowed
	if( !selected->hasWindows() )
	{
		if( a != Settings::Launch &&
			a != Settings::ShowNext )
			return;

		a = Settings::Launch;
	}

	switch( a )
	{
		case Settings::Launch:
			run( selected->getIcon()->getCommand() );
			break;
		case Settings::ShowNext:
			WindowManager::activate(
				app->getDisplay(),
				selected->getNextWindow() );
			break;
		case Settings::ShowPrevious:
			WindowManager::activate(
				app->getDisplay(),
				selected->getPreviousWindow() );
			break;
		case Settings::Hide:
			WindowManager::iconify(
				app->getDisplay(),
				selected->getNextWindow() );
			break;
		case Settings::Close:
			WindowManager::close(
				app->getDisplay(),
				selected->getNextWindow() );
			break;
	}
}

/**
 * Run some command
 *
 * @param command - command to execute
 */
int Menu::run( std::string command ) const
{
	int pid = fork();

	if( pid < 0 )
		throw "Fork failed !";
	else if( pid )
		return pid;

	char *shell = getenv( "SHELL" );

	if( !shell )
		shell = "/bin/sh";

    setsid();
    execl( shell, shell, "-c", command.c_str(), 0 );

	throw "Exec failed !";

	// make compiler happy
	return 0;
}

/**
 * Initialize menu item
 *
 * @param i - icon
 * @param s - true if sticky (optional)
 */
Menu::MenuItem::MenuItem( Icon *i, bool s ) :
	icon( i ),
	sticky( s ),
	last( 0 )
{
}

/**
 * Add a window to menu item
 *
 * @param d - display
 * @param w - window to add
 */
void Menu::MenuItem::addWindow( Display *d, Window w )
{
	XWindowAttributes xwa;

	XGetWindowAttributes( d, w, &xwa );

	WindowAttribute wa = {
		w,
		xwa.x,
		xwa.y,
		xwa.x+xwa.width,
		xwa.y+xwa.height,
		WindowManager::getWorkspace( d, w ),
		xwa.map_state };

	windowAttributes.push_back( wa );
}

/**
 * Return next window
 */
const Window Menu::MenuItem::getNextWindow()
{
	if( !hasWindows() )
		return 0;

	for( WindowAttributes::iterator i = windowAttributes.begin();
		i != windowAttributes.end();
		i++ )
		if( (*i).window == last )
		{
			if( ++i == windowAttributes.end() )
				i = windowAttributes.begin();

			last = (*i).window;

			return last;
		}

	last = (*windowAttributes.begin()).window;

	return last;
}

/**
 * Return previous window
 */
const Window Menu::MenuItem::getPreviousWindow()
{
	if( !hasWindows() )
		return 0;

	for( WindowAttributes::iterator i = windowAttributes.begin();
		i != windowAttributes.end();
		i++ )
		if( (*i).window == last )
		{
			if( i == windowAttributes.begin() )
				last = windowAttributes.back().window;
			else
				last = (*--i).window;

			return last;
		}

	last = windowAttributes.back().window;

	return last;
}

/**
 * Returns true if there's at least one window visible
 */
const bool Menu::MenuItem::isUnmapped()
{
	for( WindowAttributes::iterator i = windowAttributes.begin();
		i != windowAttributes.end();
		i++ )
		if( (*i).state != IsUnmapped )
			return false;

	return true;
}
