/*
 * Copyright (c) 2007 Markus Fisch <mf@markusfisch.de>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Markus Fisch nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#include "PieMenu.h"

#include <math.h>

using namespace PieDock;

const double PieMenu::radiansPerCircle = M_PI+M_PI;

/**
 * Initialize menu
 *
 * @param a - application
 * @param s - surface to draw menu into
 */
PieMenu::PieMenu( Application *a, Surface &s ) :
	Menu( a ),
	blender( s ),
	size( (s.getWidth() < s.getHeight() ? s.getWidth() : s.getHeight()) ),
	maxRadius( (size-static_cast<int>( .3*size ))>>1 ),
	radius( size>>2 ),
	twist( .0 ),
	centerX( s.getWidth()>>1 ),
	centerY( s.getHeight()>>1 )
{
	if( maxRadius < radius )
		throw "window too small !";
}

/**
 * Reset and update menu
 */
void PieMenu::update()
{
	radius = static_cast<int>(
		app->getSettings().getStartRadius()*maxRadius );
	twist = 0;
	lastX = lastY = -1;

	Menu::update();
}

/**
 * Returns true if menu needs to be redrawn
 *
 * @param x - x coordinate of cursor position within target surface
 * @param y - y coordinate of cursor position within target surface
 */
bool PieMenu::isObsolete( int x, int y )
{
	// menu is always obsolete while animating
	if( radius < maxRadius )
		return true;

	// redraw only if position changed
	if( x != lastX ||
		y != lastY )
	{
		lastX = x;
		lastY = y;

		return true;
	}

	return false;
}

/**
 * Draw menu
 *
 * @param x - x coordinate of cursor position within target surface
 * @param y - y coordinate of cursor position within target surface
 */
void PieMenu::draw( int x, int y )
{
	int numberOfIcons = menuItemsVector.size();
	int closestIcon = 0;
	struct
	{
		int x;
		int y;
		double weight;
		double size;
		double cellSize;
	} iconGeometries[numberOfIcons];

	// reset selected icon
	selected = 0;

	// don't do anything if there are no icons
	if( !numberOfIcons )
		return;

	// calculate positions and sizes
	{
		double circumference = M_PI*(radius<<1);
		double pixelsPerRadian = radiansPerCircle/circumference;
		int centeredY = y-centerY;
		int centeredX = x-centerX;
		double cursorAngle = atan2( centeredY, centeredX );
		double cellSize = radiansPerCircle/numberOfIcons;
		double closestAngle = 0;
		double weight = 0;
		double maxIconSize = .8*radius;
		double maxWeight;

		// calculate weight of each icon
		{
			double cursorRadius = sqrt(
				(centeredY*centeredY)+(centeredX*centeredX) );
			double infieldRadius = radius>>1;
			double f = cursorRadius/infieldRadius;

			if( cursorRadius < infieldRadius )
			{
				double b = (circumference/numberOfIcons)*.75;

				if( b < maxIconSize )
					maxIconSize = b+(maxIconSize-b)*f;
			}

			// determine how close every icon is to the cursor
			{
				double closestDistance = radiansPerCircle;
				double a = twist;
				double m = (maxIconSize*pixelsPerRadian)/cellSize;

				maxWeight = M_PI_2+pow( M_PI, m );

				for( int n = 0; n < numberOfIcons; n++ )
				{
					double d = fabs( getAngleDifference( a, cursorAngle ) );

					if( d < closestDistance )
					{
						closestDistance = d;
						closestIcon = n;
						closestAngle = a;
					}

					if( cursorRadius < infieldRadius )
						d *= f;

					iconGeometries[n].weight = M_PI_2+pow( M_PI-d, m );
					weight += iconGeometries[n].weight;

					if( (a += cellSize) > M_PI )
						a -= radiansPerCircle;
				}
			}
		}

		// calculate size of icons
		{
			double sizeUnit = circumference/weight;

			for( int n = numberOfIcons; n--; )
				iconGeometries[n].size =
					iconGeometries[n].cellSize =
						sizeUnit*iconGeometries[n].weight;

			// scale icons within cell
			{
				double maxSize = sizeUnit*maxWeight;

				if( maxSize > maxIconSize )
				{
					double f = maxIconSize/maxSize;

					for( int n = numberOfIcons; n--; )
						iconGeometries[n].size *= f;
				}
			}
		}

		// calculate icon positions
		{
			double difference = getAngleDifference(
				cursorAngle, closestAngle );
			double angle = getValidAngle(
				cursorAngle-
					(pixelsPerRadian*
						iconGeometries[closestIcon].cellSize)/cellSize*
					difference );

			// active icon
			iconGeometries[closestIcon].x =
				centerX+static_cast<int>( radius*cos( angle ) );
			iconGeometries[closestIcon].y =
				centerY+static_cast<int>( radius*sin( angle ) );

			// calculate positions of all other icons
			{
				double leftAngle = angle;
				double rightAngle = angle;
				int left = closestIcon;
				int right = closestIcon;
				int previousRight = closestIcon;
				int previousLeft = closestIcon;

				for( int n = 0; ; n++ )
				{
					if( (--left) < 0 )
						left = numberOfIcons-1;

					// break here when number of icons is odd
					if( right == left )
						break;

					if( (++right) >= numberOfIcons )
						right = 0;

					leftAngle = getValidAngle(
						leftAngle-
							(
								(.5*iconGeometries[previousLeft].cellSize)+
								(.5*iconGeometries[left].cellSize)
							)*pixelsPerRadian );

					iconGeometries[left].x =
						centerX+static_cast<int>( radius*cos( leftAngle ) );
					iconGeometries[left].y =
						centerY+static_cast<int>( radius*sin( leftAngle ) );

					// break here when number of icons is even
					if( left == right )
						break;

					rightAngle = getValidAngle(
						rightAngle+
							(
								(.5*iconGeometries[previousRight].cellSize)+
								(.5*iconGeometries[right].cellSize)
							)*pixelsPerRadian );

					iconGeometries[right].x =
						centerX+static_cast<int>( radius*cos( rightAngle ) );
					iconGeometries[right].y =
						centerY+static_cast<int>( radius*sin( rightAngle ) );

					previousRight = right;
					previousLeft = left;
				}
			}
		}
	}

	// draw icons
	{
		const Settings::ActiveIndicator *activeIndicator =
			&app->getSettings().getActiveIndicator();
		RgbaSurfaceSizeMap *activeIndicatorSizeMap =
			activeIndicator->getSizeMap();
		int n = 0;

		for( MenuItemsVector::iterator i = menuItemsVector.begin();
			i != menuItemsVector.end();
			i++, n++ )
		{
			const int size = static_cast<int>( iconGeometries[n].size )>>1<<1;
			const RgbaSurface *surface =
				(*i)->getIcon()->getSurface( size, size );
			const int x = iconGeometries[n].x-(surface->getWidth()>>1);
			const int y = iconGeometries[n].y-(surface->getHeight()>>1);
			const int opacity =
				(n == closestIcon ?
					app->getSettings().getFocusedAlpha() :
					app->getSettings().getUnfocusedAlpha());

			blender.blend(
				*surface,
				x,
				y,
				opacity );

			if( (*i)->hasWindows() )
			{
				const int activeIndicatorSize = size/3;
				const RgbaSurface *s =
					(activeIndicatorSizeMap ?
						activeIndicatorSizeMap :
						(*i)->getIcon())->getSurface(
							activeIndicatorSize,
							activeIndicatorSize );

				struct
				{
					int left;
					int top;
					int right;
					int bottom;
				} spot;

				spot.left = x+activeIndicator->getX(
					activeIndicatorSize,
					surface->getWidth() );
				spot.top = y+activeIndicator->getY(
					activeIndicatorSize,
					surface->getHeight() );
				spot.right = spot.left+s->getWidth();
				spot.bottom = spot.top+s->getHeight();

				blender.blend(
					*s,
					spot.left,
					spot.top,
					opacity );
			}

			if( n == closestIcon )
				selected = *i;
		}
	}

	// zoom and rotate into appearance
	if( radius < maxRadius )
	{
		if( (radius += 2) > maxRadius )
			radius = maxRadius;

		if( (twist += .05) > radiansPerCircle )
			twist -= radiansPerCircle;
	}
}

/**
 * Turn menu
 *
 * @param r - radians to turn
 */
void PieMenu::turn( double r )
{
	twist += r;

	if( twist > radiansPerCircle )
		twist -= radiansPerCircle;
	else if( twist < 0 )
		twist += radiansPerCircle;

	// force redraw
	lastX = lastY = -1;
}
