/*
 * Copyright (c) 2007 Markus Fisch <mf@markusfisch.de>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Markus Fisch nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#include "PieMenuWindow.h"
#include "XUtil.h"

using namespace PieDock;

/**
 * Initialize this window
 *
 * @param a - Application object
 */
PieMenuWindow::PieMenuWindow( Application &a ) :
	TransparentWindow( a ),
	menu( &a, canvas )
{
	XSelectInput( 
		app->getDisplay(), 
		window,
		LeaveWindowMask |
			ButtonReleaseMask | 
			ButtonPressMask );
}

/**
 * Make window appear
 */
void PieMenuWindow::appear()
{
	menu.update();
	show();
}

/**
 * (Re-)Draw contents
 */
void PieMenuWindow::draw()
{
	XUtil::Position p = XUtil::getCursorPosition( 
		app->getDisplay(),
		window );

	if( !menu.isObsolete( p.x, p.y ) )
		return;

	clear();
	menu.draw( p.x, p.y );
	update();
}

/**
 * Process event, this method returns true as long as it wants to keep 
 * processing events
 *
 * @param event - X event
 */
bool PieMenuWindow::processEvent( XEvent &event )
{
	switch( event.type )
	{
		case LeaveNotify:
			hide();
			return false;
		case ButtonRelease:
			{
				Settings::ButtonFunctions *bf = 
					&app->getSettings().getButtonFunctions();

				for( Settings::ButtonFunctions::iterator i = bf->begin();
					i != bf->end();
					i++ )
					if( (*i).button == event.xbutton.button )
						switch( (*i).action )
						{
							case Settings::Launch:
							case Settings::ShowNext:
							case Settings::ShowPrevious:
							case Settings::Hide:
							case Settings::Close:
								hide();
								menu.execute( (*i).action );
								return false;
							case Settings::SpinUp:
								menu.turn( .1 );
								break;
							case Settings::SpinDown:
								menu.turn( -.1 );
								break;
							case Settings::Disappear:
								hide();
								return false;
							default:
								menu.execute( (*i).action );
								break;
						}
			}
			break;
	}

	return true;
}

/**
 * Show window
 */
void PieMenuWindow::show()
{
	centerUnderCursor();
	TransparentWindow::show();
}

/**
 * Center window under mouse cursor
 */
void PieMenuWindow::centerUnderCursor() const
{
	int screen = DefaultScreen( app->getDisplay() );
	int desktopWidth = DisplayWidth( app->getDisplay(), screen );
	int desktopHeight = DisplayHeight( app->getDisplay(), screen );
	XUtil::Position p = XUtil::getCursorPosition( app->getDisplay() );

	p.x -= width>>1;
	p.y -= height>>1;

	// shift window a little bit left to avoid an unpleasant 
	// flickering while appearing
	p.x -= width>>3;

	if( p.x < 0 )
		p.x = 0;
	else if( p.x+width >= desktopWidth )
		p.x = desktopWidth-width;

	if( p.y < 0 )
		p.y = 0;
	else if( p.y+height >= desktopHeight )
		p.y = desktopHeight-height;

	XMoveWindow( app->getDisplay(), window, p.x, p.y );
}
