/*
 * Copyright (c) 2007 Markus Fisch <mf@markusfisch.de>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Markus Fisch nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#include "PngSurface.h"

#include <fstream>

using namespace PieDock;

/**
 * Initialize surface from file
 *
 * @param f - path and filename of PNG to load
 */
PngSurface::PngSurface( std::string f ) :
	RgbaSurface( 0, 0 )
{
	std::ifstream in( f.c_str(), std::ios::in );

	if( !in.good() )
		throw ("Cannot open PNG file \""+f+"\" !").c_str();

	load( in );
}

/**
 * Initialize surface from stream
 *
 * @param in - input stream
 */
PngSurface::PngSurface( std::istream &in ) :
	RgbaSurface( 0, 0 )
{
	load( in );
}

/**
 * Load image from a PNG input stream
 *
 * @param in - input stream
 */
void PngSurface::load( std::istream &in )
{
	png_structp png = 0;
	png_infop info = 0;

	if( !(png = png_create_read_struct( 
			PNG_LIBPNG_VER_STRING,
			0, 0, 0 )) )
		throw "PNG library error !";

	if( !(info = png_create_info_struct( png )) ||
		setjmp( png_jmpbuf( png ) ) ||
		!in.good() )
		throw "Cannot open read from PNG stream !";

	png_set_read_fn( png, reinterpret_cast<void *>( &in ), readPng );

	png_set_sig_bytes( png, 0 );
	png_read_png( png, info, 
		PNG_TRANSFORM_IDENTITY | 
		PNG_TRANSFORM_EXPAND |
		PNG_TRANSFORM_BGR,
		0 );

	if( !(info->valid & PNG_INFO_IDAT) )
		throw "Invalid PNG stream !";

	freeData();
	calculateSize( info->width, info->height );
	allocateData();

	for( int y = 0, *src, *dest = reinterpret_cast<int *>( getData() ); 
		y < info->height && 
			(src = reinterpret_cast<int *>( info->row_pointers[y] )); 
		y++, dest += info->width )
		memcpy( dest, src, info->width<<2 );

	// convert grayscale image to rgba, png_set_gray_to_rgb() doesn't work
	// for some reason
	if( info->color_type == PNG_COLOR_TYPE_GRAY_ALPHA )
		for( int y = info->height, 
				*line = reinterpret_cast<int *>( getData() );
			--y; 
			line += info->width )
		{
			unsigned char *dest = 
				reinterpret_cast<unsigned char *>( line )+
				(info->width<<2);
			unsigned char *src = 
				reinterpret_cast<unsigned char *>( line )+
				(info->width<<1);

			for( int x = info->width; --x; )
			{
				*(--dest) = *(--src);
				*(--dest) = *(--src);
				*(--dest) = *src;
				*(--dest) = *src;
			}
		}

	png_destroy_read_struct( &png, &info, (png_infopp) 0 );
}

/**
 * Read png data from a stream; this needs to be a static member
 *
 * @param png - png pointer
 * @param data - pointer to data
 * @param length - data length
 */
void PngSurface::readPng( 
	png_structp png, 
	png_bytep data, 
	png_size_t length )
{
	std::istream *in = reinterpret_cast<std::istream *>( png->io_ptr );

	if( !in ||
		!in->good() )
	{
		png_error( png, "Cannot read from stream" );
		return;
	}

	in->read( reinterpret_cast<char *>( data ), length );
}
