/*
 * Copyright (c) 2007 Markus Fisch <mf@markusfisch.de>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Markus Fisch nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#include "Settings.h"
#include "Application.h"

#include <sys/types.h>
#include <dirent.h>
#include <fstream>
#include <sstream>

using namespace PieDock;

/**
 * Initialize settings
 *
 * @param f - path and name of resource file
 */
Settings::Settings( std::string f ) :
	rcFile( f )
{
}

/**
 * (Re-)Load settings
 */
void Settings::load()
{
	std::ifstream in( rcFile.c_str(), std::ios::in );

	if( !in ||
		!in.good() )
		return;

	// initialize temporary memorizers
	typedef std::vector<std::pair<std::string, std::string> > StickyCandidates;

	enum PreloadSetting
	{
		PreloadNone = 0,
		PreloadAll,
		PreloadSticky
	};

	PreloadSetting preload = PreloadNone;
	StickyCandidates stickyCandidates;

	// reset settings to default values
	{
		width = height = 256;
		keys.clear();
		buttons.clear();
		buttonFunctions.clear();
		iconMap.reset();
		stickyIcons.clear();
		activeIndicator.reset();
		focusedAlpha = unfocusedAlpha = 0xff;
		startRadius = .9;
	}

	// parse file
	Statement statement;

	for( int line = 1; getline( in, statement ); line++ )
	{
		// cut off comments
		{
			std::string::size_type p;

			if( (p = statement.find( '#' )) != std::string::npos )
				statement.erase( ++p );
		}

		// skip empty lines
		if( statement.empty() )
			continue;

		// split statement into tokens
		std::vector<std::string> tokens = statement.split();
		std::vector<std::string>::iterator i = tokens.begin();

		if( !(*i).compare( "size" ) )
		{
			if( ++i == tokens.end() )
				parsingError(
					"Insufficient arguments to size directive",
					line );
			else
				width = height = atoi( (*i).c_str() );
		}
		else if( !(*i).compare( "width" ) )
		{
			if( ++i == tokens.end() )
				parsingError(
					"Insufficient arguments to width directive",
					line );
			else
				width = atoi( (*i).c_str() );
		}
		else if( !(*i).compare( "height" ) )
		{
			if( ++i == tokens.end() )
				parsingError(
					"Insufficient arguments to width directive",
					line );
			else
				height = atoi( (*i).c_str() );
		}
		else if( !(*i).compare( "trigger" ) )
		{
			int n = tokens.size();

			if( n < 3 ||
				n > 4 )
				parsingError(
					"Insufficient arguments to trigger directive",
					line );
			else
			{
				std::string type = (*++i);
				struct Trigger trigger = {
					AnyModifier };

				if( n == 4 )
				{
					i++;

					if( !(*i).compare( "Mod1" ) )
						trigger.modifier = Mod1Mask;
					else if( !(*i).compare( "Mod2" ) )
						trigger.modifier = Mod2Mask;
					else if( !(*i).compare( "Mod3" ) )
						trigger.modifier = Mod3Mask;
					else if( !(*i).compare( "Mod4" ) )
						trigger.modifier = Mod4Mask;
					else if( !(*i).compare( "Mod5" ) )
						trigger.modifier = Mod5Mask;
					else if( !(*i).compare( "Control" ) )
						trigger.modifier = ControlMask;
					else if( !(*i).compare( "Shift" ) )
						trigger.modifier = ShiftMask;
					else if( !(*i).compare( "Lock" ) )
						trigger.modifier = LockMask;
				}

				if( !type.compare( "key" ) )
				{
					trigger.keySym = XStringToKeysym( (*++i).c_str() );
					keys.push_back( trigger );
				}
				else if( !type.compare( "button" ) )
				{
					trigger.button = atoi( (*++i).c_str() );
					buttons.push_back( trigger );
				}
				else
					parsingError(
						"Unknown trigger type",
						line );
			}
		}
		else if( !(*i).compare( "button" ) )
		{
			if( tokens.size() != 3 )
				parsingError(
					"Insufficient arguments to button directive",
					line );
			else
			{
				ButtonFunction bf = { 0, Settings::NoAction };

				switch( atoi( (*++i).c_str() ) )
				{
					case 1:
						bf.button = Button1;
						break;
					case 2:
						bf.button = Button2;
						break;
					case 3:
						bf.button = Button3;
						break;
					case 4:
						bf.button = Button4;
						break;
					case 5:
						bf.button = Button5;
						break;
					case 6:
						bf.button = 6; // not defined by X11
						break;
					case 7:
						bf.button = 7; // not defined by X11
						break;
					default:
						parsingError(
							"Invalid button number",
							line );
						break;
				}

				if( bf.button )
				{
					i++;

					if( !(*i).compare( "Launch" ) )
						bf.action = Launch;
					else if( !(*i).compare( "ShowNext" ) )
						bf.action = ShowNext;
					else if( !(*i).compare( "ShowPrevious" ) )
						bf.action = ShowPrevious;
					else if( !(*i).compare( "Hide" ) )
						bf.action = Hide;
					else if( !(*i).compare( "Close" ) )
						bf.action = Close;
					else if( !(*i).compare( "SpinUp" ) )
						bf.action = SpinUp;
					else if( !(*i).compare( "SpinDown" ) )
						bf.action = SpinDown;
					else if( !(*i).compare( "Disappear" ) )
						bf.action = Disappear;
					else
						parsingError(
							"Invalid action",
							line );

					if( bf.action != Settings::NoAction )
						buttonFunctions.push_back( bf );
				}
			}			
		}
		else if( !(*i).compare( "path" ) )
		{
			if( ++i == tokens.end() )
				parsingError(
					"Insufficient arguments to path directive",
					line );
			else
			{
				std::string path = (*i);

				// expand ~
				if( path[0] == '~' )
					path = Application::getHome()+path.substr( 1 );

				// ensure there is a terminating slash
				if( path[path.length()-1] != '/' )
					path += '/';

				iconMap.setPath( path );
			}
		}
		else if( !(*i).compare( "alias" ) )
		{
			switch( tokens.size() )
			{
				default:
					parsingError(
						"Insufficient arguments to alias directive",
						line );
					break;
				case 3:
					{
						// can't do this in just one line since the 
						// compiler will iterate the latter one first,
						// at least with g++ 4.1.1
						std::string a = *++i;
						std::string n = *++i;

						iconMap.addNameAlias( a, n );
					}
					break;
				case 4:
					{
						std::string t = *++i;
						std::string a = *++i;
						std::string n = *++i;

						if( !t.compare( "name" ) )
							iconMap.addNameAlias( a, n );
						else if( !t.compare( "class" ) )
							iconMap.addClassAlias( a, n );
						else if( !t.compare( "title" ) )
							iconMap.addTitleAlias( a, n );					
						else
							parsingError(
								"Unknown alias type",
								line );
					}
					break;
			}
		}
		else if( !(*i).compare( "sticky" ) )
		{
			if( ++i == tokens.end() )
				parsingError(
					"Insufficient arguments to sticky directive",
					line );
			else
			{
				std::string name = *i;
				std::string command;

				if( ++i != tokens.end() )
					command = *i;

				stickyCandidates.push_back(
					std::make_pair( name, command ) );
			}
		}
		else if( !(*i).compare( "preload" ) )
		{
			if( ++i == tokens.end() )
				parsingError(
					"Insufficient arguments to preload directive",
					line );
			else
			{
				if( !(*i).compare( "all" ) )
					preload = PreloadAll;
				else if( !(*i).compare( "sticky" ) )
					preload = PreloadSticky;
				else if( !(*i).compare( "none" ) )
					preload = PreloadNone;
				else
					parsingError(
						"Unknown argument to preload directive",
						line );
			}
		}
		else if( !(*i).compare( "active-indicator" ) )
		{
			if( ++i == tokens.end() )
				parsingError(
					"Insufficient arguments to active-indicator directive",
					line );
			else
			{
				// load file
				{
					std::string file = (*i);

					// expand ~
					if( file[0] == '~' )
						file = Application::getHome()+file.substr( 1 );

					activeIndicator.load( file );
				}

				if( ++i != tokens.end() )
				{
					activeIndicator.parseX( (*i) );

					if( ++i != tokens.end() )
						activeIndicator.parseY( (*i) );
				}
			}
		}
		else if( !(*i).compare( "focused" ) )
		{
			if( ++i == tokens.end() )
				parsingError(
					"Insufficient arguments to focused directive",
					line );
			else
				focusedAlpha = atoi( (*i).c_str() ) % 0xff;
		}
		else if( !(*i).compare( "unfocused" ) )
		{
			if( ++i == tokens.end() )
				parsingError(
					"Insufficient arguments to unfocused directive",
					line );
			else
				unfocusedAlpha = atoi( (*i).c_str() ) % 0xff;
		}
		else if( !(*i).compare( "start-radius" ) )
		{
			if( ++i == tokens.end() )
				parsingError(
					"Insufficient arguments to start-radius directive",
					line );
			else if( (startRadius = atof( (*i).c_str() )) > 1 )
				startRadius = 1;
		}
	}

	// sticky icons should be realized after parsing the whole file
	// to ensure all alias- and path-directives are processed
	for( StickyCandidates::iterator i = stickyCandidates.begin();
		i != stickyCandidates.end();
		i++ )
	{
		Icon *icon;

		if( !(icon = iconMap.getIconByName( (*i).first )) )
			icon = iconMap.getFallbackIcon( (*i).first );

		if( icon )
		{
			if( !(*i).second.empty() )
				icon->setCommand( (*i).second );

			stickyIcons.push_back( icon );
		}
	}

	// this should be done after parsing the whole file to ensure
	// all alias- and path-directives are processed
	if( preload != PreloadNone )
	{
		// those values are guesses for use with a PieMenu; if that
		// application supports more menu forms some day, this must
		// be done individually
		int size = (width > height ? width : height);
		int min = 2;
		int max = (static_cast<int>( size*.280 )>>1)<<1;
		int step = 2;

		switch( preload )
		{
			case PreloadAll:
				{
					DIR *d = opendir( iconMap.getPath().c_str() );

					if( !d )
						throw "Cannot open image directory !";

					for( struct dirent *e; (e = readdir( d )); )
					{
						strtok( e->d_name, "." );

						if( !e->d_name ||
							*e->d_name == '.' )
							continue;

						Icon *icon = iconMap.getIconByName( e->d_name );

						if( !icon )
							continue;

						presizeIcon(
							icon,
							min,
							min,
							max,
							max,
							step,
							step );
					}

					closedir( d );
				}
				break;
			case PreloadSticky:
				for( StickyCandidates::iterator i = stickyCandidates.begin();
					i != stickyCandidates.end();
					i++ )
				{
					Icon *icon = iconMap.getIconByName( (*i).first );

					if( !icon )
						continue;

					presizeIcon(
						icon,
						min,
						min,
						max,
						max,
						step,
						step );
				}
				break;
		}
	}
}

/**
 * Handle a parsing error
 *
 * @param message - error message
 * @param line - line number
 */
void Settings::parsingError( const char *message, unsigned int line )
{
	std::ostringstream s;

	s << message << " in line " << line;

	throw s.str().c_str();
}

/**
 * Presize icon
 *
 * @param icon - icon to presize
 * @param fromWidth - width in pixels of smallest size
 * @param fromHeight - height in pixels of smallest size
 * @param toWidth - width in pixels of biggest size
 * @param toHeight - height in pixels of biggest size
 * @param xStep - horizontal step in pixels
 * @param yStep - vertical step in pixels
 */
void Settings::presizeIcon( Icon *icon, int fromWidth, int fromHeight,
	int toWidth, int toHeight, int xStep, int yStep )
{
	for( int w = fromWidth, h = fromHeight; 
		w <= toWidth && h <= toHeight; )
	{
		icon->getSurface( w, h );

		if( w <= toWidth )
			w += xStep;

		if( h <= toHeight )
			h += yStep;
	}
}

/**
 * Split statement into its tokens
 *
 * @param delimiter - characters on which to split (optional)
 */
std::vector<std::string> Settings::Statement::split( const char *delimiter )
{
	std::vector<std::string> v;
	std::string::size_type last = 0;
	std::string::size_type until; 

	for( ;
		(until = lengthUntil( delimiter, last )) != std::string::npos;
		last = ++until )
		v.push_back( trim( substr( last, until-last ) ) );

	v.push_back( trim( substr( last ) ) );

	return v;
}

/**
 * Calculate the length of the initial segment of this string
 * which consists entirely of characters not in reject; respect
 * quoting
 *
 * @param reject - character that may appear within the span
 * @param offset - offset in string (optional)
 */
std::string::size_type Settings::Statement::lengthUntil( 
	const char *reject,
	std::string::size_type offset )
{
	for( const char *s = c_str()+offset; *s; s++ )
	{
		if( *s == '\\' )
		{
			if( !*(++s) )
				break;

			continue;
		}

		if( *s == '"' )
		{
			for( ++s;
				 *s && *s != '"';
				 s++ )
				if( *s == '\\' )
				{
					if( !*(++s) )
						break;

					continue;
				}

			if( !*s )
				break;
		}

		for( const char *r = reject; *r; r++ )
			if( *r == *s )
				return s-c_str();
	}

	return std::string::npos;
}

/**
 * Return trimed string
 *
 * @param s - string to trim
 * @param d - characters to trim (optional)
 */
std::string Settings::Statement::trim( std::string s, const char *d )
{
	size_type p;

	if( (p = s.find_first_not_of( d )) != 
		std::string::npos )
		s.erase( 0, p );

	if( (p = s.find_last_not_of( d )) != 
		std::string::npos )
		s.erase( ++p );

	return s;
}
