/*
 * Copyright (c) 2007 Markus Fisch <mf@markusfisch.de>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Markus Fisch nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef _PieDock_Settings_
#define _PieDock_Settings_

#include "IconMap.h"
#include "PngSurface.h"

#include <X11/Xlib.h>
#include <vector>
#include <string>

namespace PieDock
{
	/**
	 * Settings container
	 *
	 * @author Markus Fisch <mf@markusfisch.de>
	 */
	class Settings
	{
		public:
			enum Action
			{
				NoAction = 0,
				Launch = 1,
				ShowNext = 2,
				ShowPrevious = 3,
				Hide = 4,
				Close = 5,
				SpinUp = 6,
				SpinDown = 7,
				Disappear = 8
			};

			typedef struct Trigger
			{
				unsigned int modifier;
				union
				{
					unsigned int button;
					KeySym keySym;
				};
			} Key, Button;

			typedef struct
			{
				unsigned int button;
				Action action;
			} ButtonFunction;

			/**
			 * ActiveIndicator
			 */
			class ActiveIndicator
			{
				public:
					ActiveIndicator() : sizeMap( 0 ) { reset(); }
					virtual ~ActiveIndicator() { clear(); }
					inline RgbaSurfaceSizeMap *getSizeMap() const { 
						return sizeMap; }
					inline void load( std::string f ) 
					{
						clear();
						PngSurface s( f );
						sizeMap = new RgbaSurfaceSizeMap( s );
					}
					inline void parseX( std::string s )
					{
						if( !s.compare( "left" ) )
							xAlign = Left;
						else if( !s.compare( "center" ) )
							xAlign = Center;
						else if( !s.compare( "right" ) )
							xAlign = Right;
					}
					inline void parseY( std::string s )
					{
						if( !s.compare( "top" ) )
							yAlign = Top;
						else if( !s.compare( "center" ) )
							yAlign = Center;
						else if( !s.compare( "bottom" ) )
							yAlign = Bottom;
					}
					inline const int getX( int s, int i ) const
					{
						switch( xAlign )
						{
							default:
							case Left:
								return 0;
							case Center:
								return (i-s)>>1;
							case Right:
								return i-s;
						}
					}
					inline const int getY( int s, int i ) const
					{
						switch( yAlign )
						{
							default:
							case Top:
								return 0;
							case Center:
								return (i-s)>>1;
							case Bottom:
								return i-s;
						}
					}
					inline void reset()
					{ 
						clear(); 
						xAlign = Right;
						yAlign = Bottom;
					}

				private:
					enum Alignment
					{
						Left,
						Right,
						Top,
						Bottom,
						Center
					};

					RgbaSurfaceSizeMap *sizeMap;
					Alignment xAlign;
					Alignment yAlign;

					inline void clear() { delete sizeMap; sizeMap = 0; }
			};

			typedef std::vector<Key> Keys;
			typedef std::vector<Button> Buttons;
			typedef std::vector<ButtonFunction> ButtonFunctions;
			typedef std::vector<Icon *> Icons;

			Settings( std::string );
			virtual ~Settings() {}
			inline const int getWidth() const { return width; }
			inline const int getHeight() const { return height; }
			inline Keys &getKeys() { return keys; }
			inline Buttons &getButtons() { return buttons; }
			inline ButtonFunctions &getButtonFunctions() { 
				return buttonFunctions; }
			inline IconMap &getIconMap() { return iconMap; }
			inline Icons &getStickyIcons() { return stickyIcons; }
			inline ActiveIndicator &getActiveIndicator() { 
				return activeIndicator; }
			inline const int &getFocusedAlpha() const { return focusedAlpha; }
			inline const int &getUnfocusedAlpha() const { 
				return unfocusedAlpha; }
			inline const double &getStartRadius() const {
				return startRadius; }
			virtual void load();

		protected:
			std::string rcFile;
			int width;
			int height;
			Keys keys;
			Buttons buttons;
			ButtonFunctions buttonFunctions;
			IconMap iconMap;
			Icons stickyIcons;
			ActiveIndicator activeIndicator;
			int focusedAlpha;
			int unfocusedAlpha;
			double startRadius;

			virtual void parsingError( const char *, unsigned int );
			virtual void presizeIcon( Icon *, int, int, int, int, int, int );

		private:
			/**
			 * A statement string
			 */
			class Statement : public std::string
			{
				public:
					Statement( const char *s ) : std::string( s ) {}
					Statement( std::string s ) : std::string( s ) {}
					Statement() : std::string() {}
					virtual ~Statement() {}
					std::vector<std::string> split( const char * = " \t" );

				private:
					std::string::size_type lengthUntil( const char *, 
						 std::string::size_type = 0 );
					std::string trim( std::string, 
						const char * = " \t\r\n\"" );
			};
	};
}

#endif
