/*
 * Copyright (c) 2007 Markus Fisch <mf@markusfisch.de>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Markus Fisch nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#include "Surface.h"

#include <string>

using namespace PieDock;

/**
 * Copy constructor
 *
 * @param s - some surface
 */
Surface::Surface( Surface &s )
{
	width = s.getWidth();
	height = s.getHeight();
	depth = s.getDepth();
	bytesPerPixel = s.getBytesPerPixel();
	bytesPerLine = s.getBytesPerLine();
	padding = s.getPadding();
	size = s.getSize();

	allocateData();

	memcpy( data, s.getData(), size );
}

/**
 * Free resources
 */
Surface::~Surface()
{
	freeData();
}

/**
 * Initialize surface; this constructor is protected
 */
Surface::Surface() : 
	data( 0 ),
	width( 0 ),
	height( 0 ),
	depth( 0 ),
	bytesPerPixel( 0 ),
	bytesPerLine( 0 ),
	padding( 0 ),
	size( 0 )
{
}

/**
 * Calculate size in bytes of this surface
 *
 * @param w - width of surface in pixels
 * @param h - height of surface in pixels
 * @param d - color depth, bits per pixel (optional)
 */
void Surface::calculateSize( int w, int h, int d )
{
	width = w;
	height = h;
	depth = d;
	bytesPerPixel = depth>>3;

	// calculate bytes per line
	{
		int bitsPerLine = width*depth;

		// ceil to full byte if bits per pixel is fewer than a byte
		// or divide by 8 if there are at last 8 bits per pixel
		if( depth < 8 )
			bytesPerLine = (bitsPerLine+7) & ~7;
		else
			bytesPerLine = bitsPerLine>>3;
	}

	// pad every image line to a multiple of 4 bytes
	padding = bytesPerLine;
	bytesPerLine = (bytesPerLine+3) & ~3;
	padding = bytesPerLine-padding;

	size = height*bytesPerLine;
}

/**
 * Allocate data for surface
 */
void Surface::allocateData()
{
	if( !(data = new unsigned char[size]) )
		throw "Cannot allocate surface memory !";
}

/**
 * Free data of surface
 */
void Surface::freeData()
{
	// it's valid to delete 0
	delete data;
}
