/*
 * Copyright (c) 2007 Markus Fisch <mf@markusfisch.de>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Markus Fisch nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef _PieDock_WindowManager_
#define _PieDock_WindowManager_

#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <string>
#include <vector>
#include <map>

namespace PieDock
{
	/**
	 * Interface to the window manager
	 *
	 * @author Markus Fisch <mf@markusfisch.de>
	 */
	class WindowManager
	{
		public:
			/**
			 * Window list
			 */
			class WindowList : public std::vector<Window>
			{
				public:
					WindowList( Display *d ) { addClientsOf( d ); }
					virtual ~WindowList() {}
					void addClientsOf( Display * );
			};

			virtual ~WindowManager() {}
			static void activate( Display *, Window );
			static void iconify( Display *, Window );
			static void close( Display *, Window );
			static Window getActive( Display * );
			static std::string getTitle( Display *, Window );
			static unsigned long getWorkspace( Display *, Window );
			static unsigned long getNumberOfWorkspaces( Display * );
			static void sendClientMessage( Display *, Window, const char *,
				unsigned long = 0, unsigned long = 0, unsigned long = 0,
				unsigned long = 0, unsigned long = 0 );

		private:
			/**
			 * X property template
			 */
			template <class T> class Property
			{
				public:
					Property( Display *d, Window w ) :
						display( d ),
						window( w ),
						data( 0 ),
						items( 0 ) {}
					Property( Display *d, Window w, Atom type,
						const char *name ) :
						Property( d, w ) { fetch( type, name ); }
					virtual ~Property() { freeData(); }
					inline T *getData() const { return data; }
					inline unsigned long getItems() const { return items; }
					bool fetch( Atom type, const char *name,
						long length = 1024, long offset = 0,
						Bool remove = False )
					{
						freeData();

						Atom returnedType;
						int format;
						unsigned long items;
						unsigned long bytesAfter;
						unsigned char *data;

						if( XGetWindowProperty(
								display,
								window,
								getAtom( display, name ),
								offset,
								length,
								remove,
								type,
								&returnedType,
								&format,
								&items,
								&bytesAfter,
								&data ) != Success )
							return false;

						if( returnedType != type )
						{
							freeData();
							return false;
						}

						this->data = reinterpret_cast<T *>( data );
						this->items = items;

						return true;
					}

				private:
					Display *display;
					Window window;
					T *data;
					unsigned long items;

					void freeData()
					{
						if( !data )
							return;

						XFree( reinterpret_cast<unsigned char *>( data ) );

						data = 0;
						items = 0;
					}
			};

			typedef std::map<std::string, Atom> StringToAtom;
			static StringToAtom stringToAtom;

			WindowManager() {}
			static Atom getAtom( Display *, const char * );
	};
}

#endif
