/*
 * Copyright (c) 2007 Markus Fisch <mf@markusfisch.de>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Markus Fisch nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#include "XSurface.h"

#include <string>
#include <X11/Xutil.h>

using namespace PieDock;

/**
 * Initialize surface
 *
 * @param w - width of image in pixels
 * @param h - height of image in pixels
 * @param d - X Display
 */
XSurface::XSurface( int w, int h, Display *d ) :
	display( d ),
	resource( 0 )
{
	calculateSize( w, h, determineDepth() );
	allocateData();
}

/**
 * Clean up
 */
XSurface::~XSurface()
{
	freeData();
}

/**
 * Determine bits per pixel of an screen-compatible XImage, this does 
 * not need to be the same as the color-depth of the screen; when
 * creating an XImage, it may have another color-depth than the screen
 */
int XSurface::determineDepth()
{
	XPixmapFormatValues *pf;
	int formats;
	int bitsPerPixel;
	int screen = DefaultScreen( display );

	bitsPerPixel = DefaultDepth( display, screen );
	pf = XListPixmapFormats( display, &formats );

	if( formats )
	{
		int format;

		for( format = formats;
			 format--; )
			if( pf[format].depth == bitsPerPixel )
			{
				bitsPerPixel = pf[format].bits_per_pixel;
				break; 
			}

		XFree( reinterpret_cast<char *>( pf ) );
	}

	return bitsPerPixel;
}

/**
 * Allocate data
 */
void XSurface::allocateData()
{
	if( !(data = reinterpret_cast<unsigned char *>( calloc( 
			size, 
			sizeof( char ) ) )) )
		throw "Cannot allocate memory !";

	int screen = DefaultScreen( display );

	if( !(resource = XCreateImage( 
			display, 
			DefaultVisual( display, screen ), 
			DefaultDepth( display, screen ),
			ZPixmap, 
			0, 
			reinterpret_cast<char *>( data ), 
			width, 
			height, 
			32, 
			0 )) )
		throw "Cannot create XImage";
}

/**
 * Free data
 */
void XSurface::freeData()
{
	XDestroyImage( resource );
	resource = 0;
	// data is freed by XDestroyImage
	data = 0;
}
