/*
 * Copyright (c) 2007 Markus Fisch <mf@markusfisch.de>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Markus Fisch nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#include "Application.h"

#include <sys/wait.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <libgen.h>
#include <iostream>

bool stop = false;

/**
 * Signal handler
 *
 * @param id - signal id
 */
void signalHandler( int id )
{
	switch( id )
	{
		case SIGCHLD:
			// more than one process may have terminated
        	while( waitpid( -1, 0, WNOHANG | WUNTRACED ) > 0 );
        	break;
		case SIGHUP:
		case SIGINT:
		case SIGTERM:
			stop = true;
			break;
	}

	return;
}

/**
 * Process trivial command line arguments
 *
 * @param argc - number of arguments
 * @param argv - pointer to pointers of arguments
 */
void processArguments( int argc, char **argv )
{
	for( char *binary = *argv; --argc; )
		if( **++argv == '-' )
			switch( *((*argv)+1) )
			{
				default:
					std::cerr << "Skipping unknown flag '" <<
						*((*argv)+1) << "'" << std::endl;
					break;
				case '?':
				case 'h':
				case 'v':
					std::cout << 
						basename( binary ) << " 0.1.7" << 
						" (c) 2007 Markus Fisch <mf@markusfisch.de>" << 
						std::endl;
					exit( 0 );
			}
		else
			std::cerr << "Skipping unknown argument \"" << 
				*argv << "\"" << std::endl;
}

/**
 * Process entry
 *
 * @param argc - number of arguments
 * @param argv - pointer to pointers of arguments
 */
int main( int argc, char **argv )
{
	try
	{
		// process trivial arguments
		processArguments( argc, argv );

		switch( fork() )
		{
			default:
				// terminate parent process to detach from shell
				return 0;
			case 0:
				// pursue in child process
				break;
			case -1:
				throw "Cannot fork";
		}

		// always open display after fork
		PieDock::Application a( argc, argv );

		// if another instance is already running, wake it
		if( a.wake() )
			return 0;

		// obtain new process group
		setsid();

		signal( SIGCHLD, signalHandler );
		signal( SIGHUP, signalHandler );
		signal( SIGINT, signalHandler );
		signal( SIGTERM, signalHandler );

		return a.run( &stop );
	}
	catch( const char *e )
	{
		std::cerr << e << std::endl;

		return -1;
	}
}
